@@ -2,7 +2,7 @@
 """
 Wrapper for libsodium library
 
-Copyright (c) 2013, Marsiske Stefan and individual contributors.
+Copyright (c) 2013-2014, Marsiske Stefan.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modification,
@@ -27,157 +27,172 @@ OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 """
 
-from pysodium import sodium
+import ctypes
 
-crypto_box_NONCEBYTES = sodium.lib.crypto_box_NONCEBYTES
-crypto_box_PUBLICKEYBYTES = sodium.lib.crypto_box_PUBLICKEYBYTES
-crypto_box_SECRETKEYBYTES = sodium.lib.crypto_box_SECRETKEYBYTES
-crypto_box_ZEROBYTES = sodium.lib.crypto_box_ZEROBYTES
-crypto_box_BOXZEROBYTES = sodium.lib.crypto_box_BOXZEROBYTES
-crypto_secretbox_KEYBYTES = sodium.lib.crypto_secretbox_KEYBYTES
-crypto_secretbox_NONCEBYTES = sodium.lib.crypto_secretbox_NONCEBYTES
-crypto_secretbox_KEYBYTES = sodium.lib.crypto_secretbox_KEYBYTES
-crypto_secretbox_ZEROBYTES = sodium.lib.crypto_box_ZEROBYTES
-crypto_secretbox_BOXZEROBYTES = sodium.lib.crypto_box_BOXZEROBYTES
-crypto_sign_PUBLICKEYBYTES = sodium.lib.crypto_sign_PUBLICKEYBYTES
-crypto_sign_SECRETKEYBYTES = sodium.lib.crypto_sign_SECRETKEYBYTES
-crypto_stream_KEYBYTES = sodium.lib.crypto_stream_KEYBYTES
-crypto_stream_NONCEBYTES = sodium.lib.crypto_stream_NONCEBYTES
-crypto_generichash_BYTES = sodium.lib.crypto_generichash_BYTES
-crypto_scalarmult_curve25519_BYTES = sodium.lib.crypto_scalarmult_curve25519_BYTES
-crypto_scalarmult_BYTES = sodium.lib.crypto_scalarmult_curve25519_BYTES
-crypto_sign_BYTES = sodium.lib.crypto_sign_BYTES
+sodium = ctypes.cdll.LoadLibrary("libsodium.so")
+crypto_box_NONCEBYTES = 24L
+crypto_box_PUBLICKEYBYTES = 32L
+crypto_box_SECRETKEYBYTES = 32L
+crypto_box_ZEROBYTES = 32L
+crypto_box_BOXZEROBYTES = 16L
+crypto_secretbox_KEYBYTES = 32L
+crypto_secretbox_NONCEBYTES = 24L
+crypto_secretbox_KEYBYTES = 32L
+crypto_secretbox_ZEROBYTES = 32L
+crypto_secretbox_BOXZEROBYTES = 16L
+crypto_sign_PUBLICKEYBYTES = 32L
+crypto_sign_SECRETKEYBYTES = 64L
+crypto_stream_KEYBYTES = 32L
+crypto_stream_NONCEBYTES = 24L
+crypto_generichash_BYTES = 32L
+crypto_scalarmult_curve25519_BYTES = 32L
+crypto_scalarmult_BYTES = 32L
+crypto_sign_BYTES = 64L
+
+"""
+typedef struct crypto_generichash_blake2b_state {
+    uint64_t h[8];
+    uint64_t t[2];
+    uint64_t f[2];
+    uint8_t  buf[256];
+    size_t   buflen;
+    uint8_t  last_node;
+    ...;
+} crypto_generichash_state;
+"""
+crypto_generichash_state = 357L
 
 def crypto_scalarmult_curve25519(n,p):
-    buf = sodium.ffi.new("unsigned char[]", crypto_scalarmult_BYTES)
-    sodium.lib.crypto_scalarmult_curve25519(buf, n, p)
-    return sodium.ffi.buffer(buf, crypto_scalarmult_BYTES)[:]
+    buf = ctypes.create_string_buffer(crypto_scalarmult_BYTES)
+    sodium.crypto_scalarmult_curve25519(buf, n, p)
+    return buf.raw
 
 def crypto_scalarmult_curve25519_base(n):
-    buf = sodium.ffi.new("unsigned char[]", crypto_scalarmult_BYTES)
-    sodium.lib.crypto_scalarmult_curve25519_base(buf, n)
-    return sodium.ffi.buffer(buf, crypto_scalarmult_BYTES)[:]
+    buf = ctypes.create_string_buffer( crypto_scalarmult_BYTES)
+    sodium.crypto_scalarmult_curve25519_base(buf, n)
+    return buf.raw
 
+
+# crypto_generichash(unsigned char *out, size_t outlen, const unsigned char *in, unsigned long long inlen, const unsigned char *key, size_t keylen)
 def crypto_generichash(m, k=b'', outlen=crypto_generichash_BYTES):
-    # FIXME returns different result than the 3-step procedure used as a workaround
-    buf = sodium.ffi.new("unsigned char[]", outlen)
-    sodium.lib.crypto_generichash(buf, len(buf), m, len(m), k, len(k))
-    return sodium.ffi.buffer(buf, outlen)[:]
+    buf = ctypes.create_string_buffer(outlen)
+    if not sodium.crypto_generichash(buf, ctypes.c_uint(outlen), m, ctypes.c_ulonglong(len(m)), k, ctypes.c_uint(len(k))) == 0:
+        raise ValueError
+    return buf.raw
 
 #crypto_generichash_init(crypto_generichash_state *state, const unsigned char *key, const size_t keylen, const size_t outlen);
 def crypto_generichash_init(outlen=crypto_generichash_BYTES, k=b''):
-    buf = sodium.ffi.new("crypto_generichash_state[]", 1)
-    sodium.lib.crypto_generichash_init(buf, k, len(k), outlen)
-    return buf
+    buf = ctypes.create_string_buffer(crypto_generichash_state)
+    sodium.crypto_generichash_init(buf, k, ctypes.c_ulonglong(len(k)), outlen)
+    return buf.raw
 
 #crypto_generichash_update(crypto_generichash_state *state, const unsigned char *in, unsigned long long inlen);
 def crypto_generichash_update(state, m):
-    sodium.lib.crypto_generichash_update(state, m, len(m))
+    sodium.crypto_generichash_update(state, m, ctypes.c_ulonglong(len(m)))
     return state
 
 #crypto_generichash_final(crypto_generichash_state *state, unsigned char *out, const size_t outlen);
 def crypto_generichash_final(state, outlen=crypto_generichash_BYTES):
-    buf = sodium.ffi.new("unsigned char[]", outlen)
-    sodium.lib.crypto_generichash_final(state, buf, outlen )
-    return sodium.ffi.buffer(buf, outlen)[:]
+    buf = ctypes.create_string_buffer(outlen)
+    sodium.crypto_generichash_final(state, buf, outlen)
+    return buf.raw
 
-def randombytes(l):
-    buf = sodium.ffi.new("unsigned char[]", l)
-    sodium.lib.randombytes(buf, l)
-    return sodium.ffi.buffer(buf, l)[:]
+def randombytes(size):
+    buf = ctypes.create_string_buffer(size)
+    sodium.randombytes(buf, ctypes.c_ulonglong(size))
+    return buf.raw
 
 def crypto_box_keypair():
-    pk = sodium.ffi.new("unsigned char[]", crypto_box_PUBLICKEYBYTES)
-    sk = sodium.ffi.new("unsigned char[]", crypto_box_SECRETKEYBYTES)
-    if not sodium.lib.crypto_box_keypair(pk, sk):
+    pk = ctypes.create_string_buffer( crypto_box_PUBLICKEYBYTES)
+    sk = ctypes.create_string_buffer( crypto_box_SECRETKEYBYTES)
+    if not sodium.crypto_box_keypair(pk, sk) == 0:
         raise ValueError
-    pk = sodium.ffi.buffer(pk, crypto_box_PUBLICKEYBYTES)[:]
-    sk = sodium.ffi.buffer(sk, crypto_box_SECRETKEYBYTES)[:]
-    return (pk, sk)
+    return (pk.raw, sk.raw)
 
 def crypto_box(msg, nonce, pk, sk):
     if None in (msg, nonce, pk, sk): raise ValueError
     padded = b"\x00" * crypto_box_ZEROBYTES + msg
-    c = sodium.ffi.new("unsigned char[]", len(padded))
-    if not sodium.lib.crypto_box(c, padded, len(padded), nonce, pk, sk):
+    c = ctypes.create_string_buffer( len(padded))
+    if not sodium.crypto_box(c, padded, ctypes.c_ulonglong(len(padded)), nonce, pk, sk) == 0:
         raise ValueError
-    return sodium.ffi.buffer(c, len(padded))[crypto_box_BOXZEROBYTES:]
+    return c.raw[crypto_box_BOXZEROBYTES:]
 
 def crypto_box_open(c, nonce, pk, sk):
     if None in (c, nonce, pk, sk): raise ValueError
     padded = b"\x00" * crypto_box_BOXZEROBYTES + c
-    msg = sodium.ffi.new("unsigned char[]", len(padded))
-    if not sodium.lib.crypto_box_open(msg, padded, len(padded), nonce, pk, sk):
+    msg = ctypes.create_string_buffer( len(padded))
+    if not sodium.crypto_box_open(msg, padded, ctypes.c_ulonglong(len(padded)), nonce, pk, sk) == 0:
         raise ValueError
-    return sodium.ffi.buffer(msg, len(padded))[crypto_box_ZEROBYTES:]
+    return msg.raw[crypto_box_ZEROBYTES:]
 
 def crypto_secretbox(msg, nonce, k):
     if None in (msg, nonce, k): raise ValueError
     padded = b"\x00" * crypto_secretbox_ZEROBYTES + msg
-    c = sodium.ffi.new("unsigned char[]", len(padded))
-    if not sodium.lib.crypto_secretbox(c, padded, len(padded), nonce, k):
+    c = ctypes.create_string_buffer( len(padded))
+    if not sodium.crypto_secretbox(c, padded, ctypes.c_ulonglong(len(padded)), nonce, k) == 0:
         raise ValueError
-    return sodium.ffi.buffer(c, len(padded))[crypto_secretbox_BOXZEROBYTES:]
+    return c.raw[crypto_secretbox_BOXZEROBYTES:]
 
 def crypto_secretbox_open(c, nonce, k):
     if None in (c, nonce, k): raise ValueError
     padded = b"\x00" * crypto_secretbox_BOXZEROBYTES + c
-    msg = sodium.ffi.new("unsigned char[]", len(padded))
-    if not sodium.lib.crypto_secretbox_open(msg, padded, len(padded), nonce, k):
+    msg = ctypes.create_string_buffer( len(padded))
+    if not sodium.crypto_secretbox_open(msg, padded, ctypes.c_ulonglong(len(padded)), nonce, k) == 0:
         raise ValueError
-    return sodium.ffi.buffer(msg, len(padded))[crypto_secretbox_ZEROBYTES:]
+    return msg.raw[crypto_secretbox_ZEROBYTES:]
 
 def crypto_sign_keypair():
-    pk = sodium.ffi.new("unsigned char[]", crypto_sign_PUBLICKEYBYTES)
-    sk = sodium.ffi.new("unsigned char[]", crypto_sign_SECRETKEYBYTES)
-    if not sodium.lib.crypto_sign_keypair(pk, sk):
+    pk = ctypes.create_string_buffer(crypto_sign_PUBLICKEYBYTES)
+    sk = ctypes.create_string_buffer(crypto_sign_SECRETKEYBYTES)
+    if not sodium.crypto_sign_keypair(pk, sk) == 0:
         raise ValueError
-    pk = sodium.ffi.buffer(pk, crypto_sign_PUBLICKEYBYTES)[:]
-    sk = sodium.ffi.buffer(sk, crypto_sign_SECRETKEYBYTES)[:]
-    return (pk, sk)
+    return (pk.raw, sk.raw)
 
 def crypto_sign(m, sk):
     if None in (m, sk): raise ValueError
-    smsg = sodium.ffi.new("unsigned char[]", len(m)+crypto_sign_BYTES)
-    smsglen = sodium.ffi.new("unsigned long long *")
-    if not sodium.lib.crypto_sign(smsg, smsglen, m, len(m), sk):
+    smsg = ctypes.create_string_buffer(len(m)+crypto_sign_BYTES)
+    smsglen = ctypes.pointer(ctypes.c_ulonglong())
+    if not sodium.crypto_sign(smsg, smsglen, m, ctypes.c_ulonglong(len(m)), sk) == 0:
         raise ValueError
-    return sodium.ffi.buffer(smsg, smsglen[0])[:]
+    return smsg.raw
 
 def crypto_sign_open(sm, pk):
     if None in (sm, pk): raise ValueError
-    msg = sodium.ffi.new("unsigned char[]", len(sm))
-    msglen = sodium.ffi.new("unsigned long long *")
-    if not sodium.lib.crypto_sign_open(msg, msglen, sm, len(sm), pk):
+    msg = ctypes.create_string_buffer(len(sm))
+    msglen = ctypes.c_ulonglong()
+    msglenp = ctypes.pointer(msglen)
+    if not sodium.crypto_sign_open(msg, msglenp, sm, ctypes.c_ulonglong(len(sm)), pk) == 0:
         raise ValueError
-    return sodium.ffi.buffer(msg, msglen[0])[:]
+    return msg.raw[:msglen.value]
 
 def crypto_stream(cnt, nonce = None, key = None):
-    res = sodium.ffi.new("unsigned char[]", cnt)
+    res = ctypes.create_string_buffer(cnt)
     if not nonce:
         nonce = randombytes(crypto_stream_NONCEBYTES)
     if not key:
         key = randombytes(crypto_stream_KEYBYTES)
-    if not sodium.lib.crypto_stream(res, len(res), nonce, key):
+    if not sodium.crypto_stream(res, ctypes.c_ulonglong(cnt), nonce, key) == 0:
         raise ValueError
-    return sodium.ffi.buffer(res, cnt)[:]
+    return res.raw
 
 def crypto_stream_xor(msg, cnt, nonce = None, key = None):
-    res = sodium.ffi.new("unsigned char[]", cnt)
+    res = ctypes.create_string_buffer(cnt)
     if not nonce:
         nonce = randombytes(crypto_stream_NONCEBYTES)
     if not key:
         key = randombytes(crypto_stream_KEYBYTES)
-    if not sodium.lib.crypto_stream_xor(res, msg, len(res), nonce, key):
+    if not sodium.crypto_stream_xor(res, msg, ctypes.c_ulonglong(cnt), nonce, key) == 0:
         raise ValueError
-    return sodium.ffi.buffer(res, cnt)[:]
+    return res.raw
 
 def test():
     import binascii
 
-    print(binascii.hexlify(crypto_stream(8)))
-    print(binascii.hexlify(crypto_stream(16)))
-    print(binascii.hexlify(crypto_stream(32)))
+    crypto_stream(8L)
+    crypto_stream(1337L)
+    print(binascii.hexlify(crypto_stream(8L)))
+    print(binascii.hexlify(crypto_stream(16L)))
+    print(binascii.hexlify(crypto_stream(32L)))
     print(binascii.hexlify(crypto_stream_xor('howdy', len('howdy'))))
     print(binascii.hexlify(crypto_stream_xor('howdy' * 16, len('howdy')*16)))
 
@@ -208,7 +223,12 @@ def test():
 
     pk, sk = crypto_sign_keypair()
     signed = crypto_sign('howdy',sk)
-    print(crypto_sign_open(signed, pk))
+    changed = signed[:crypto_sign_BYTES]+'0'+signed[crypto_sign_BYTES+1:]
+    print crypto_sign_open(signed, pk)
+    try:
+        crypto_sign_open(changed, pk)
+    except ValueError:
+        print "signature failed to verify for changed payload"
 
 if __name__ == '__main__':
     test()
